//See the file LICENSE for redistribution information.

//Copyright (c) 2002,2007 Oracle.  All rights reserved.

//$Id: XmlDocument.java,v 1.86 2008/03/25 17:14:49 gmf Exp $

package com.sleepycat.dbxml;

import java.util.Arrays;
import java.io.InputStream;
import java.util.Collections;
import java.util.Vector;

import com.sleepycat.dbxml.XmlValue.NodeValue;

public class XmlDocument {
    protected Vector metaData;
    protected long docID;
    protected int cid;
    protected Content content;
    /*
     * State information for the document, such as its transaction or temporary
     * database, is stored either in an XmlResults or in an XmlManager with an
     * optional XmlTransaction. The XmlResults is set if the document is created
     * from an XmlResults or XmlValue, the XmlManager is set if the document is
     * created from an XmlContainer, XmlManager, or XmlValue.
     */
    protected XmlManager manager;
    protected XmlTransaction txn;
    protected XmlResults results;

    //Used by the function putDocumentAsEventWriter
    private long eventWriter;

    public XmlDocument() {
	metaData = new Vector();
	docID = 0;
	cid = 0;
	content = new Content();
    }

    public XmlDocument(XmlDocument o) throws XmlException {
	docID = o.docID;
	cid = o.cid;
	metaData = new Vector(o.metaData);
	content = new Content(o.content);
    }

    //Deprecated. 
    public void delete() {}

    public void setContentAsEventReader(XmlEventReader reader)
    throws XmlException {
	content.setContentAsEventReader(reader);
    }

    public InputStream getContentAsInputStream() throws XmlException {
	return new XmlInputStreamWrap(getContentAsXmlInputStream());
    }

    public void setContentAsXmlInputStream(XmlInputStream adopted)
    throws XmlException {
	content.setContentAsXmlInputStream(adopted);
    }

    public boolean getMetaData(String uri, String name, XmlValue value)
    throws XmlException {
	for (int i = 0; i < metaData.size(); i++) {
	    XmlMetaData data = (XmlMetaData) metaData.get(i);
	    if (uri.matches(data.get_uri()) && name.matches(data.get_name())
		    && !data.get_removed()) {
		XmlValue.setValue(value, data.get_value());
		return true;
	    }
	}
	XmlValue returned = HelperFunctions.getMetaData(uri, name, this);
	if(!returned.isNull()){
	    XmlValue.setValue(value, returned);
	    return true;
	}
	return false;
    }

    public void setName(String name) throws XmlException {
	String newName = name;
	if(newName == null)
	    newName = "";
	XmlValue value = new XmlValue(newName);
	setMetaData(dbxml_javaConstants.metaDataNamespace_uri,
		dbxml_javaConstants.metaDataName_name, value);
    }

    public String getName() throws XmlException {
	XmlValue value = new XmlValue();
	boolean exists = getMetaData(dbxml_javaConstants.metaDataNamespace_uri,
		dbxml_javaConstants.metaDataName_name, value);
	if (exists)
	    return value.asString();
	return "";
    }

    public void setContent(String content) throws XmlException {
	this.content.setContent(content);
    }

    public void setContent(byte[] content) throws XmlException {
	this.content.setContent(content);
    }

    public XmlInputStream getContentAsXmlInputStream() throws XmlException {
	if (content.isStream()){
	    XmlInputStream ins = content.getXmlInputStream();
	    content.stream = null;
	    content.type = NONE;
	    return ins;
	}else if (!content.hasContent() && docID == 0)
	    return null;
	else
	    return HelperFunctions.getContentAsXmlInputStream(this);
    }

    public XmlEventReader getContentAsEventReader() throws XmlException {
	if (content.isReader()){
	    XmlEventReader xer = content.getEventReader();
	    content.reader = null;
	    content.type = NONE;
	    return xer;
	}else if (!content.hasContent() && docID == 0)
	    return null;
	else
	    return HelperFunctions.getContentAsEventReader(this);
    }

    public void getContentAsEventWriter(XmlEventWriter writer)
    throws XmlException {
	try{
	    HelperFunctions.getContentAsEventWriter(this, writer);
	}catch(XmlException e){
	    writer.release();
	    throw e;
	}
	writer.release();
    }

    public void fetchAllData() 
    throws XmlException {
	materializeMetaData();
	getContent();
    }

    public void setMetaData(String uri, String name, XmlValue value)
    throws XmlException {
	addMetaData(uri, name, value, true, false);
    }

    public void removeMetaData(String uri, String name) throws XmlException {
	XmlValue temp = new XmlValue("x");
	addMetaData(uri, name, temp, false, true);
    }

    public void setMetaData(String uri, String name, byte[] value)
    throws XmlException {
	XmlValue data = new XmlValue(XmlValue.BINARY, value);
	addMetaData(uri, name, data, true, false);
    }

    public String getContentAsString() throws XmlException {
	if (content.isDBT())
	    return content.getString();
	else if (!content.hasContent() && docID == 0)
	    return "";
	else{
	    String results = HelperFunctions.getContentAsString(this);
	    content.setContent(results);
	    return results;
	}
    }

    public byte[] getContent() throws XmlException {
	if (content.isDBT())
	    return content.getDBT();
	else if (!content.hasContent() && docID == 0){
	    byte [] empty = {};
	    return empty;
	}else{
	    byte [] results = HelperFunctions.getContent(this);
	    content.setContent(results);
	    return results;
	}
    }

    public byte[] getMetaData(String uri, String name) throws XmlException {
	XmlValue data = new XmlValue();
	if(getMetaData(uri, name, data)){
	    if(data.isBinary())
		return data.asBinary();
	    return data.asString().getBytes();
	}
	return null;
    }

    public XmlMetaDataIterator getMetaDataIterator() throws XmlException {
	materializeMetaData();
	return new XmlMetaDataIterator(this);
    }

    public boolean equals(XmlDocument other) throws XmlException {
	if(docID != 0 && docID == other.docID && cid == other.cid)
	    return true;
	else if(docID == 0 && other.docID == 0 && content == content 
		&& metaData == other.metaData)
	    return true;
	return false;
    }

    //The rest of this class is for internal use.
    /* If both modified and removed are set to false then the meta data is 
     * only being added if it does not already exist.
     */
    protected void addMetaData(String uri, String name, XmlValue value, boolean modified, boolean removed)
    throws XmlException {
	XmlMetaData data = null;
	for (int i = 0; i < metaData.size(); i++) {
	    data = (XmlMetaData)metaData.get(i);
	    if (uri.matches(data.get_uri()) && name.matches(data.get_name())){
		if(modified || removed){
		    data.set_modified(modified);
		    data.set_removed(removed);
		    data.set_value(value);
		}
		return;
	    }
	}
	data = new XmlMetaData(uri, name, value);
	if(modified || removed){
	    data.set_modified(modified);
	    data.set_removed(removed);
	}
	metaData.add(data);
    }

    protected XmlMetaData getMetaData(int index) {
	if(index >= metaData.size())
	    return null;
	return (XmlMetaData) metaData.get(index);
    }

    protected void materializeMetaData() 
    throws XmlException {
	if(docID != 0){
	    dbxml_javaJNI.materializeMetaData(this);
	}
    }

    protected XmlDocument(long documentId, int containerId) {
	metaData = new Vector();
	docID = documentId;
	cid = containerId;
	content = new Content();
    }

    protected void finalize() throws XmlException {
	metaData.clear();
	content.clear();
    }

    protected void copy(XmlDocument o) throws XmlException {
	docID = o.docID;
	cid = o.cid;
	content.clear();
	content = new Content(o.content);
	metaData.clear();
	metaData = new Vector(o.metaData);
	results = o.results;
	manager = o.manager;
	txn = o.txn;
    }

    protected void copyID(XmlDocument o) throws XmlException {
	docID = o.docID;
	cid = o.cid;
	results = null;
	manager = null;
	txn = null;
	content.clear();
	metaData.clear();
    }

    protected void clearID() throws XmlException{
	docID = 0;
	cid = 0;
    }

    protected void clearContent() throws XmlException{
	content.clear();
    }

    protected long getResultPtr() throws XmlException {
	return XmlResults.getCPtr(results);
    }

    protected long getManagerPtr() {
	return XmlManager.getCPtr(manager);
    }

    protected long getTransactionPtr() {
	return XmlTransaction.getCPtr(txn);
    }

    protected long getEventReaderPtr() throws XmlException{
	long ptr = XmlEventReader.getCPtr(content.getEventReader());
	content.reader.release();
	content.reader = null;
	content.type = NONE;
	return ptr;
    }

    protected long getInputStreamPtr() throws XmlException{
	long ptr = XmlInputStream.getCPtr(content.getXmlInputStream());
	content.stream = null;
	content.type = NONE;
	return ptr;
    }

    protected long getEventWriterPtr(){
	return eventWriter;
    }

    protected byte[] getDBT(){
	return content.dbt;
    }

    protected long getId(){
	return docID;
    }

    protected int getCid(){
	return cid;
    }

    protected int getType(){
	return content.type;
    }

    protected void setManager(XmlManager mgr)
    {
	manager = mgr;
    }

    protected void setTransaction(XmlTransaction transaction)
    {
	txn = transaction;
    }

    protected void setResults(XmlResults result)
    {
	results = result;
    }

    protected void setEventWriter(long writer){
	eventWriter = writer;
    }

    class Content {
	public XmlEventReader reader;
	public XmlInputStream stream;
	public byte[] dbt;
	public int type;

	public Content() {
	    type = NONE;
	}

	public Content(Content c) {
	    type = c.type;
	    reader = c.reader;
	    stream = c.stream;
	    dbt = c.dbt;
	}

	protected void clear() throws XmlException {
	    if (reader != null) {
		reader.close();
		reader = null;
	    }
	    if (stream != null) {
		stream.delete();
		stream = null;
	    }
	    if (dbt != null)
		dbt = null;
	    type = NONE;
	}

	public boolean hasContent() {
	    return type != NONE;
	}

	public boolean isStream() {
	    return type == INPUTSTREAM;
	}

	public boolean isReader() {
	    return type == READER;
	}

	public boolean isDBT() {
	    return type == DBT;
	}

	public void setContentAsEventReader(XmlEventReader rdr)
	throws XmlException {
	    clear();
	    type = READER;
	    reader = rdr;
	}

	public void setContentAsXmlInputStream(XmlInputStream adopted)
	throws XmlException {
	    clear();
	    type = INPUTSTREAM;
	    stream = adopted;
	}

	public void setContent(String content) throws XmlException {
	    clear();
	    type = DBT;
	    dbt = content.getBytes();
	}

	public void setContent(byte[] content) throws XmlException {
	    clear();
	    type = DBT;
	    dbt = content;
	}

	public XmlEventReader getEventReader(){
	    return reader;
	}

	public XmlInputStream getXmlInputStream(){
	    return stream;
	}

	public byte[] getDBT() {
	    return dbt;
	}

	public String getString() {
	    return new String(dbt);
	}
    }

    public final static int NONE = 0;
    public final static int DBT = 1;
    public final static int INPUTSTREAM = 2;
    public final static int DOM = 4;
    public final static int READER = 8;
    public final static int VALIDATE = 1;
    public final static int DONT_VALIDATE = 2;
    public final static int WF_ONLY = 3;
}
